@@ -2,7 +2,7 @@
 |   Copyright (C) 1999  Jochen C. Loewer (loewerj@hotmail.com)
 +----------------------------------------------------------------------------
 |
-|   $Header: /usr/local/pubcvs/tdom/generic/dom.h,v 1.28 2003/04/09 19:56:08 rolf Exp $
+|   $Id$
 |
 |
 |   A DOM interface upon the expat XML parser for the C language
@@ -74,6 +74,18 @@
 # endif /* TCL_MEM_DEBUG */
 #endif /* USE_NORMAL_ALLOCATOR */
 
+#if defined(TCL_MEM_DEBUG) || defined(NS_AOLSERVER) 
+   static void* my_malloc(size_t size){return Tcl_Alloc(size);}
+   static void  my_free(void *ptr){Tcl_Free((char*)ptr);}
+   static void* my_realloc(void *ptr,size_t size){return Tcl_Realloc(ptr,size);}
+   static XML_Memory_Handling_Suite memsuite = {
+       my_malloc, my_realloc, my_free
+   };
+#  define MEM_SUITE &memsuite
+#else
+#  define MEM_SUITE NULL
+#endif
+
 /*
  * Beginning with 8.4, Tcl API is CONST'ified
  */
@@ -81,12 +93,22 @@
 # define CONST84
 #endif
 
+/*
+ * Beginning with 8.6, interp->errorLine isn't public visible anymore
+ * (TIP 330)
+ */
+#if (TCL_MAJOR_VERSION == 8) && (TCL_MINOR_VERSION < 6)
+# define Tcl_GetErrorLine(interp) (interp)->errorLine
+#endif
+
 /*
  * Starting with Tcl 8.2 the Tcl_Panic() is defined properly
  * over the stubs table.
+ * Also, we have a proper Tcl_GetString() shortcut afterwards.
  */
 #if (TCL_MAJOR_VERSION == 8) && (TCL_MINOR_VERSION < 2)
 # define Tcl_Panic panic
+# define Tcl_GetString(a) Tcl_GetStringFromObj((a), NULL)
 #endif
 
 #define domPanic(msg) Tcl_Panic((msg));
@@ -97,29 +119,26 @@
  * way we name Tcl object accessor commands.
  */
 #ifndef TCL_THREADS
-  extern unsigned int domUniqueNodeNr;
-  extern unsigned int domUniqueDocNr;
-  extern Tcl_HashTable tagNames;
-  extern Tcl_HashTable attrNames;
+  extern unsigned long domUniqueNodeNr;
+  extern unsigned long domUniqueDocNr;
+  extern Tcl_HashTable tdom_tagNames;
+  extern Tcl_HashTable tdom_attrNames;
 # define TDomNotThreaded(x) x
 # define TDomThreaded(x)
 # define HASHTAB(doc,tab)   tab
 # define NODE_NO(doc)       ++domUniqueNodeNr
 # define DOC_NO(doc)        ++domUniqueDocNr
-# define DOC_CMD(s,doc)     sprintf((s), "domDoc%d", (doc)->documentNumber)
-# define XSLT_CMD(s,doc)    sprintf((s), "XSLTcmd%d", (doc)->documentNumber)
 #else
 # define TDomNotThreaded(x)
 # define TDomThreaded(x)    x
 # define HASHTAB(doc,tab)   (doc)->tab
 # define NODE_NO(doc)       ((doc)->nodeCounter)++
-# define DOC_NO(doc)        (unsigned int)(doc)
-# define DOC_CMD(s,doc)     sprintf((s), "domDoc0x%x", (doc)->documentNumber)
-# define XSLT_CMD(s,doc)    sprintf((s), "XSLTcmd0x%x", (doc)->documentNumber)
+# define DOC_NO(doc)        (unsigned long)(doc)
 #endif /* TCL_THREADS */
 
-# define NODE_CMD(s,node)   sprintf((s), "domNode0x%x", (unsigned int)(node))
-
+#define DOC_CMD(s,doc)      sprintf((s), "domDoc%p", (doc))
+#define NODE_CMD(s,node)    sprintf((s), "domNode%p", (node))
+#define XSLT_CMD(s,doc)     sprintf((s), "XSLTcmd%p", (doc))
 
 #define XML_NAMESPACE "http://www.w3.org/XML/1998/namespace"
 #define XMLNS_NAMESPACE "http://www.w3.org/2000/xmlns"
@@ -141,7 +160,8 @@
 #define UTF8_CHAR_LEN(c) \
   UTF8_1BYTE_CHAR((c)) ? 1 : \
    (UTF8_2BYTE_CHAR((c)) ? 2 : \
-     (UTF8_3BYTE_CHAR((c)) ? 3 : 0))
+     (UTF8_3BYTE_CHAR((c)) ? 3 : \
+       (UTF8_4BYTE_CHAR((c)) ? 4 : 0)))
 #endif
 
 /* The following 2 defines are out of the expat code */
@@ -217,6 +237,12 @@ We need 8 bits to index into pages, 3 bits to add to that index and
          : 1) \
       : 1)) \
     
+/* This definition is lax in the sense, that it accepts every 4 byte
+ * utf-8 character beyond #xFFFF as valid, no matter, if Unicode has
+ * (so far) defined a character for that encoding point. Additionally,
+ * this define does not care about the discouraged characters beyond
+ * #xFFFF (but after all, they are only discouraged, not
+ * forbidden). */
 #  define UTF8_XMLCHAR(p, n) \
   ((n) == 1 \
   ? CharBit[(int)(*(p))] \
@@ -224,7 +250,8 @@ We need 8 bits to index into pages, 3 bits to add to that index and
     ? 1 \
     : ((n) == 3 \
       ? (UTF8_XMLCHAR3(p)) \
-      : 0)))
+      : ((n) == 4 \
+        ? 1 : 0))))
 #endif
 
 #include "../expat/nametab.h"
@@ -328,37 +355,19 @@ static const unsigned char CharBit[] = {
 
 
 #if TclOnly8Bits == 1
-#  define isNameStart(x) (isalpha(*x) || ((*x)=='_') || ((*x)==':'))
-#  define isNameChar(x) (isalnum(*x)  || ((*x)=='_') || ((*x)=='-') || ((*x)=='.') || ((*x)==':'))
+#  define isNameStart(x)   (isalpha(*x) || ((*x)=='_') || ((*x)==':'))
+#  define isNameChar(x)    (isalnum(*x)  || ((*x)=='_') || ((*x)=='-') || ((*x)=='.') || ((*x)==':'))
 #  define isNCNameStart(x) (isalpha(*x) || ((*x)=='_'))
-#  define isNCNameChar(x) (isalnum(*x)  || ((*x)=='_') || ((*x)=='-') || ((*x)=='.'))
+#  define isNCNameChar(x)  (isalnum(*x)  || ((*x)=='_') || ((*x)=='-') || ((*x)=='.'))
 #else
-static int isNameStart(char *c)
-{
-    int clen;
-    clen = UTF8_CHAR_LEN (*c);
-    return (UTF8_GET_NAME_START(c, clen));
-}
-static int isNCNameStart(char *c)
-{
-    int clen;
-    clen = UTF8_CHAR_LEN (*c);
-    return (UTF8_GET_NCNAME_START(c, clen));
-}
-static int isNameChar(char *c)
-{
-    int clen;
-    clen = UTF8_CHAR_LEN (*c);
-    return (UTF8_GET_NAMING_NMTOKEN(c, clen));
-}
-static int isNCNameChar(char *c)
-{
-    int clen;
-    clen = UTF8_CHAR_LEN (*c);
-    return (UTF8_GET_NAMING_NCNMTOKEN(c, clen));
-}
+#  define isNameStart(x)   UTF8_GET_NAME_START((x),UTF8_CHAR_LEN(*(x)))
+#  define isNCNameStart(x) UTF8_GET_NCNAME_START((x),UTF8_CHAR_LEN(*(x)))
+#  define isNameChar(x)    UTF8_GET_NAMING_NMTOKEN((x),UTF8_CHAR_LEN(*(x)))
+#  define isNCNameChar(x)  UTF8_GET_NAMING_NCNMTOKEN((x),UTF8_CHAR_LEN(*(x)))
 #endif
 
+#define IS_XML_WHITESPACE(c)  ((c)==' ' || (c)=='\n' || (c)=='\r' || (c)=='\t')
+
 /*--------------------------------------------------------------------------
 |   DOMString
 |
@@ -367,9 +376,28 @@ typedef char* domString;   /* should 16-bit unicode character !!*/
 
 
 /*--------------------------------------------------------------------------
-|   DOM_nodeType
+|   domNodeType
 |
 \-------------------------------------------------------------------------*/
+#if defined(_AIX) 
+#    define    ELEMENT_NODE                 1
+#    define    ATTRIBUTE_NODE               2
+#    define    TEXT_NODE                    3
+#    define    CDATA_SECTION_NODE           4
+#    define    ENTITY_REFERENCE_NODE        5
+#    define    ENTITY_NODE                  6
+#    define    PROCESSING_INSTRUCTION_NODE  7
+#    define    COMMENT_NODE                 8
+#    define    DOCUMENT_NODE                9
+#    define    DOCUMENT_TYPE_NODE           10
+#    define    DOCUMENT_FRAGMENT_NODE       11
+#    define    NOTATION_NODE                12
+#    define    ALL_NODES                    100
+
+#    define    domNodeType                  int
+
+#else 
+
 typedef enum {
 
     ELEMENT_NODE                = 1,
@@ -387,11 +415,13 @@ typedef enum {
     ALL_NODES                   = 100
 } domNodeType;
 
+#endif
+
 /*--------------------------------------------------------------------------
 |   flags   -  indicating some internal features about nodes
 |
 \-------------------------------------------------------------------------*/
-typedef int domNodeFlags;
+typedef unsigned int domNodeFlags;
 
 #define HAS_LINE_COLUMN           1
 #define VISIBLE_IN_TCL            2
@@ -399,26 +429,23 @@ typedef int domNodeFlags;
 #define HAS_BASEURI               8
 #define DISABLE_OUTPUT_ESCAPING  16
 
-typedef int domAttrFlags;
+typedef unsigned int domAttrFlags;
 
 #define IS_ID_ATTRIBUTE           1
 #define IS_NS_NODE                2
 
-typedef int domDocFlags;
+typedef unsigned int domDocFlags;
 
-#define OUTPUT_DEFAULT_XML        1
-#define OUTPUT_DEFAULT_HTML       2
-#define OUTPUT_DEFAULT_TEXT       4
-#define OUTPUT_DEFAULT_UNKNOWN    8
-#define USE_8_BIT_ENCODING       16
-#define NEEDS_RENUMBERING        32
-#define DONT_FREE                64
+#define OUTPUT_DEFAULT_INDENT     1
+#define NEEDS_RENUMBERING         2
+#define DONT_FREE                 4
+#define IGNORE_XMLNS              8
 
 /*--------------------------------------------------------------------------
 |   a index to the namespace records
 |
 \-------------------------------------------------------------------------*/
-typedef int domNameSpaceIndex;
+typedef unsigned int domNameSpaceIndex;
 
 
 
@@ -443,21 +470,28 @@ typedef enum {
 } domException;
 
 /*--------------------------------------------------------------------------
-|   domDoctype
+|   domDocInfo
 |
 \-------------------------------------------------------------------------*/
-typedef struct domDoctype {
+typedef struct domDocInfo {
     
     /* 'name' is always the name of the documentElement, no struct element
        needed for this */
-    domString        publicId;
-    domString        systemId;
-    domString        internalSubset;
-    /* Currently missing, according to DOM 2: 'entities' and 'notations'.
-       Additional, this would be a good place to store additional 
-       informations about the document according xslt rec section 16. */
+    domString      publicId;
+    domString      systemId;
+    domString      internalSubset;
+    /* Currently missing, according to DOM 2: 'entities' and 'notations'. */
+    /* The following struct elements describes additional 'requested'
+       facets of the document, following the xslt rec, section 16 */
+    float          version;
+    char          *encoding;
+    int            omitXMLDeclaration;
+    int            standalone;
+    Tcl_HashTable *cdataSectionElements;
+    domString      method;
+    domString      mediaType;
     
-} domDoctype;
+} domDocInfo;
 
 /*--------------------------------------------------------------------------
 |   domDocument
@@ -468,7 +502,7 @@ typedef struct domDocument {
     domNodeType       nodeType  : 8;
     domDocFlags       nodeFlags : 8;
     domNameSpaceIndex dummy     : 16;
-    unsigned int      documentNumber;
+    unsigned long     documentNumber;
     struct domNode   *documentElement;
     struct domNode   *fragments;
 #ifdef TCL_THREADS
@@ -477,18 +511,22 @@ typedef struct domDocument {
     struct domNS    **namespaces;
     int               nsptr;
     int               nslen;
+    char            **prefixNSMappings; /* Stores doc global prefix ns
+                                           mappings for resolving of
+                                           prefixes in seletNodes expr */
 #ifdef TCL_THREADS
     unsigned int      nodeCounter;
 #endif
     struct domNode   *rootNode;
-    Tcl_HashTable     ids;
-    Tcl_HashTable     unparsedEntities;
-    Tcl_HashTable     baseURIs;
-    Tcl_Obj          *extResolver;
-    domDoctype       *doctype;
+    Tcl_HashTable    *ids;
+    Tcl_HashTable    *unparsedEntities;
+    Tcl_HashTable    *baseURIs;
+    Tcl_HashTable    *xpathCache;
+    char             *extResolver;
+    domDocInfo       *doctype;
     TDomThreaded (
-        Tcl_HashTable tagNames;        /* Names of tags found in doc */
-        Tcl_HashTable attrNames;       /* Names of tag attributes */
+        Tcl_HashTable tdom_tagNames;   /* Names of tags found in doc */
+        Tcl_HashTable tdom_attrNames;  /* Names of tag attributes */
         unsigned int  refCount;        /* # of object commands attached */
         struct _domlock *lock;          /* Lock for this document */
     )
@@ -524,9 +562,9 @@ typedef struct _domlock {
 \-------------------------------------------------------------------------*/
 typedef struct domNS {
 
-   char         *uri;
-   char         *prefix;
-   int           index;
+   char   *uri;
+   char   *prefix;
+   int     index;
 
 } domNS;
 
@@ -555,8 +593,13 @@ typedef struct domNode {
 
     domNodeType         nodeType  : 8;
     domNodeFlags        nodeFlags : 8;
+#ifdef TDOM_LESS_NS
     domNameSpaceIndex   namespace : 8;
-    int                 info      : 8;
+    unsigned int        info      : 8;
+#else
+    unsigned int        dummy     : 8;
+    unsigned int        info      : 8;    
+#endif
     unsigned int        nodeNumber;
     domDocument        *ownerDocument;
     struct domNode     *parentNode;
@@ -564,15 +607,26 @@ typedef struct domNode {
     struct domNode     *nextSibling;
 
     domString           nodeName;  /* now the element node specific fields */
+#ifndef TDOM_LESS_NS
+    domNameSpaceIndex   namespace;
+#endif
     struct domNode     *firstChild;
     struct domNode     *lastChild;
-#ifdef TCL_THREADS
-    struct domNode     *nextDeleted;
-#endif
     struct domAttrNode *firstAttr;
 
 } domNode;
 
+/*--------------------------------------------------------------------------
+|   domDeleteInfo
+|
+\-------------------------------------------------------------------------*/
+
+typedef struct domDeleteInfo {
+    domDocument * document;
+    domNode     * node;
+    Tcl_Interp  * interp;
+    char        * traceVarName;
+} domDeleteInfo;
 
 
 /*--------------------------------------------------------------------------
@@ -583,8 +637,13 @@ typedef struct domTextNode {
 
     domNodeType         nodeType  : 8;
     domNodeFlags        nodeFlags : 8;
+#ifdef TDOM_LESS_NS
     domNameSpaceIndex   namespace : 8;
-    int                 info      : 8;
+    unsigned int        info      : 8;
+#else
+    unsigned int        dummy     : 8;
+    unsigned int        info      : 8;    
+#endif
     unsigned int        nodeNumber;
     domDocument        *ownerDocument;
     struct domNode     *parentNode;
@@ -605,8 +664,13 @@ typedef struct domProcessingInstructionNode {
 
     domNodeType         nodeType  : 8;
     domNodeFlags        nodeFlags : 8;
+#ifdef TDOM_LESS_NS
     domNameSpaceIndex   namespace : 8;
-    int                 info      : 8;
+    unsigned int        info      : 8;
+#else
+    unsigned int        dummy     : 8;
+    unsigned int        info      : 8;    
+#endif
     unsigned int        nodeNumber;
     domDocument        *ownerDocument;
     struct domNode     *parentNode;
@@ -615,6 +679,9 @@ typedef struct domProcessingInstructionNode {
 
     domString           targetValue;   /* now the pi specific fields */
     int                 targetLength;
+#ifndef TDOM_LESS_NS
+    domNameSpaceIndex   namespace;
+#endif
     domString           dataValue;
     int                 dataLength;
 
@@ -629,8 +696,14 @@ typedef struct domAttrNode {
 
     domNodeType         nodeType  : 8;
     domAttrFlags        nodeFlags : 8;
+#ifdef TDOM_LESS_NS
     domNameSpaceIndex   namespace : 8;
-    int                 info      : 8;
+    unsigned int        info      : 8;
+#else
+    unsigned int        dummy     : 8;
+    unsigned int        info      : 8;    
+    domNameSpaceIndex   namespace;
+#endif
     domString           nodeName;
     domString           nodeValue;
     int                 valueLength;
@@ -650,85 +723,102 @@ typedef void (*domFreeCallback) (domNode * node, void * clientData);
 |   Function prototypes
 |
 \-------------------------------------------------------------------------*/
-char *         domException2String (domException expection);
+const char *   domException2String (domException expection);
+
 
 void           domModuleInitialize (void);
-domDocument *  domCreateDoc ();
-domDocument *  domCreateDocument (Tcl_Interp *interp,
-                                  char *documentElementTagName,
-                                  char *uri);
+domDocument *  domCreateDoc (const char *baseURI, int storeLineColumn);
+domDocument *  domCreateDocument (Tcl_Interp *interp, const char *uri,
+                                  char *documentElementTagName);
+void           domSetDocumentElement (domDocument *doc);
 
-domDocument *  domReadDocument (XML_Parser parser,
-                                char *xml,
-                                int   length,
-                                int   ignoreWhiteSpaces,
-                                TEncoding *encoding_8bit,
-                                int   storeLineColumn,
-                                int   feedbackAfter,
-                                Tcl_Channel channel,
-                                char *baseurl,
-                                Tcl_Obj *extResolver,
-                                int   useForeignDTD,
-                                int   paramEntityParsing,
-                                Tcl_Interp *interp);
+domDocument *  domReadDocument   (XML_Parser parser,
+                                  char *xml,
+                                  int   length,
+                                  int   ignoreWhiteSpaces,
+                                  TEncoding *encoding_8bit,
+                                  int   storeLineColumn,
+                                  int   ignoreXMLNS,
+                                  int   feedbackAfter,
+                                  Tcl_Obj *feedbackCmd,
+                                  Tcl_Channel channel,
+                                  const char *baseurl,
+                                  Tcl_Obj *extResolver,
+                                  int   useForeignDTD,
+                                  int   paramEntityParsing,
+                                  Tcl_Interp *interp,
+                                  int  *status);
 
-void           domFreeDocument (domDocument *doc, domFreeCallback freeCB, void * clientData);
-void           domFreeNode (domNode *node, domFreeCallback freeCB, void *clientData, int dontfree);
+void           domFreeDocument   (domDocument *doc, 
+                                  domFreeCallback freeCB, 
+                                  void * clientData);
 
-domTextNode *  domNewTextNode (domDocument *doc,
-                               char        *value,
-                               int          length,
-                               domNodeType  nodeType);
+void           domFreeNode       (domNode *node, 
+                                  domFreeCallback freeCB, 
+                                  void *clientData, 
+                                  int dontfree);
+
+domTextNode *  domNewTextNode    (domDocument *doc,
+                                  const char  *value,
+                                  int          length,
+                                  domNodeType  nodeType);
 
 domNode *      domNewElementNode (domDocument *doc,
-                                  char        *tagName,
-		                  domNodeType  nodeType);
+                                  const char  *tagName);
 		
 domNode *      domNewElementNodeNS (domDocument *doc,
-                                    char        *tagName,
-                                    char        *uri,
-		                    domNodeType  nodeType);
+                                    const char  *tagName,
+                                    const char  *uri);
 
 domProcessingInstructionNode * domNewProcessingInstructionNode (
                                   domDocument *doc,
-                                  char        *targetValue,
+                                  const char  *targetValue,
                                   int          targetLength,
-                                  char        *dataValue,
+                                  const char  *dataValue,
                                   int          dataLength);
 
-domAttrNode *  domSetAttribute (domNode *node, char *attributeName,
-                                               char *attributeValue);
+domAttrNode *  domSetAttribute (domNode *node, const char *attributeName,
+                                               const char *attributeValue);
 
-domAttrNode *  domSetAttributeNS (domNode *node, char *attributeName,
-                                                 char *attributeValue,
-                                                 char *uri,
+domAttrNode *  domSetAttributeNS (domNode *node, const char *attributeName,
+                                                 const char *attributeValue,
+                                                 const char *uri,
                                                  int   createNSIfNeeded);
-domAttrNode *  domGetAttributeNodeNS (domNode *node, char *uri, 
-                                                     char *localname);
+domAttrNode *  domGetAttributeNodeNS (domNode *node, const char *uri, 
+                                                     const char *localname);
 
-int            domRemoveAttribute (domNode *node, char *attributeName);
-int            domRemoveAttributeNS (domNode *node, char *uri, char *localName);
+int            domRemoveAttribute (domNode *node, const char *attributeName);
+int            domRemoveAttributeNS (domNode *node, const char *uri,
+                                     const char *localName);
+domNode *      domPreviousSibling (domNode *attr);
 domException   domDeleteNode   (domNode *node, domFreeCallback freeCB, void *clientData);
 domException   domRemoveChild  (domNode *node, domNode *childToRemove);
 domException   domAppendChild  (domNode *node, domNode *childToAppend);
 domException   domInsertBefore (domNode *node, domNode *childToInsert, domNode *refChild);
 domException   domReplaceChild (domNode *node, domNode *newChild, domNode *oldChild);
-domException   domSetNodeValue (domNode *node, char *nodeValue, int valueLen);
+domException   domSetNodeValue (domNode *node, const char *nodeValue,
+                                int valueLen);
 domNode *      domCloneNode (domNode *node, int deep);
 
 domTextNode *  domAppendNewTextNode (domNode *parent, char *value, int length, domNodeType nodeType, int disableOutputEscaping);
-domNode *      domAppendNewElementNode (domNode *parent, char *tagName, char *uri);
+domNode *      domAppendNewElementNode (domNode *parent, const char *tagName,
+                                        const char *uri);
 domNode *      domAppendLiteralNode (domNode *parent, domNode *node);
 domNS *        domAddNSToNode (domNode *node, domNS *nsToAdd);
-char *         domNamespacePrefix (domNode *node);
-char *         domNamespaceURI    (domNode *node);
-char *         domGetLocalName    (char *nodeName);
-int            domSplitQName (char *name, char *prefix, char **localName);
-domNS *        domLookupNamespace (domDocument *doc, char *prefix, char *namespaceURI);
-domNS *        domLookupPrefix  (domNode *node, char *prefix);
+const char *   domNamespacePrefix (domNode *node);
+const char *   domNamespaceURI    (domNode *node);
+const char *   domGetLocalName    (const char *nodeName);
+int            domSplitQName (const char *name, char *prefix,
+                              const char **localName);
+domNS *        domLookupNamespace (domDocument *doc, const char *prefix, 
+                                   const char *namespaceURI);
+domNS *        domLookupPrefix  (domNode *node, const char *prefix);
+const char *   domLookupPrefixWithMappings (domNode *node, const char *prefix,
+                                            char **prefixMappings);
 domNS *        domLookupURI     (domNode *node, char *uri);
 domNS *        domGetNamespaceByIndex (domDocument *doc, int nsIndex);
-domNS *        domNewNamespace (domDocument *doc, char *prefix, char *namespaceURI);
+domNS *        domNewNamespace (domDocument *doc, const char *prefix,
+                                const char *namespaceURI);
 int            domGetLineColumn (domNode *node, int *line, int *column);
 
 int            domXPointerChild (domNode * node, int all, int instance, domNodeType type,
@@ -751,16 +841,27 @@ int            domXPointerXSibling (domNode * node, int forward_mode, int all, i
                                     char *attrValue, int attrLen,
                                     domAddCallback addCallback, void * clientData);
 
-char *         findBaseURI (domNode *node);
+const char *   findBaseURI (domNode *node);
 
-void           tcldom_tolower (char *str, char *str_out, int  len);
-int            domIsNAME (char *name);
-int            domIsQNAME (char *name);
-int            domIsNCNAME (char *name);
-int            domIsChar (char *str);
+void           tcldom_tolower (const char *str, char *str_out, int  len);
+int            domIsNAME (const char *name);
+int            domIsPINAME (const char *name);
+int            domIsQNAME (const char *name);
+int            domIsNCNAME (const char *name);
+int            domIsChar (const char *str);
+int            domIsBMPChar (const char *str);
+int            domIsComment (const char *str);
+int            domIsCDATA (const char *str);
+int            domIsPIValue (const char *str);
 void           domCopyTo (domNode *node, domNode *parent, int copyNS);
+void           domCopyNS (domNode *from, domNode *to);
 domAttrNode *  domCreateXMLNamespaceNode (domNode *parent);
 void           domRenumberTree (domNode *node);
+int            domPrecedes (domNode *node, domNode *other);
+void           domNormalize (domNode *node, int forXPath, 
+                             domFreeCallback freeCB, void *clientData);
+domException   domAppendData (domTextNode *node, char *value, int length, 
+                              int disableOutputEscaping);
 
 #ifdef TCL_THREADS
 void           domLocksLock(domlock *dl, int how);
@@ -770,8 +871,6 @@ void           domLocksDetach(domDocument *doc);
 void           domLocksFinalize(ClientData dummy);
 #endif
 
-domDocument *  domCreateEmptyDoc(void);
-
 /*---------------------------------------------------------------------------
 |   coercion routines for calling from C++
 |
