@@ -34,28 +34,9 @@
 |
 |   ------------------------------------------------------------------------
 |
-|   A parser for XML.
-|
-|   Copyright (C) 1998 D. Richard Hipp
-|
-|   This library is free software; you can redistribute it and/or
-|   modify it under the terms of the GNU Library General Public
-|   License as published by the Free Software Foundation; either
-|   version 2 of the License, or (at your option) any later version.
-|
-|   This library is distributed in the hope that it will be useful,
-|   but WITHOUT ANY WARRANTY; without even the implied warranty of
-|   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-|   Library General Public License for more details.
-|
-|   You should have received a copy of the GNU Library General Public
-|   License along with this library; if not, write to the
-|   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-|   Boston, MA  02111-1307, USA.
-|
-|   Author contact information:
-|     drh@acm.org
-|     http://www.hwaci.com/drh/
+|   Partly based on a parser for XML (for TMML by R.Hipp 1998). 
+|   This source code is released into the public domain by the author.
+|   on 2002, December 17.
 |
 \---------------------------------------------------------------------------*/
 
@@ -79,8 +60,8 @@
 #define DBG(x)          
 #define RetError(m,p)   *errStr = tdomstrdup(m); *pos = p; return TCL_ERROR;
 #define SPACE(c)        ((c)==' ' || (c)=='\n' || (c)=='\t' || (c)=='\r')
-#define IsLetter(c)     ( ((c)>='A' && (c)<='Z') || ((c)>='a' && (c)<='z') )
-#define TU(c)           toupper(c)
+#define IsLetter(c)     ( ((c)>='A' && (c)<='Z') || ((c)>='a' && (c)<='z') || ((c) >= '0' && (c) <= '9') )
+#define TU(c)           toupper((unsigned char)c)
 
 
 
@@ -544,7 +525,7 @@ static Er er_sequences[] = {
 \---------------------------------------------------------------------------*/
 static void ErInit (void)
 {
-    int i;  /* For looping thru the list of entity references */
+    size_t i;  /* For looping thru the list of entity references */
     int h;  /* The hash on a entity */
 
     for(i=0; i<sizeof(er_sequences)/sizeof(er_sequences[0]); i++){
@@ -652,7 +633,7 @@ static void TranslateEntityRefs (
                 }
 #endif
             } else {
-                while (z[i] && isalpha(z[i])) {
+                while (z[i] && isalpha((unsigned char)z[i])) {
                    i++;
                 }
                 c = z[i];
@@ -724,7 +705,7 @@ HTML_SimpleParse (
     register char *x, *start, *piSep;
     int            saved;
     int            hasContent;
-    domNode        *pnode, *toplevel;
+    domNode       *pnode;
     domNode       *node = NULL, *parent_node = NULL;
     domTextNode   *tnode;
     domAttrNode   *attrnode, *lastAttr;
@@ -732,11 +713,6 @@ HTML_SimpleParse (
     int            only_whites   = 0;
     int            hnew, autoclose, ignore;
     char           tmp[250], *y = NULL;
-/**
- #define LATE_CLOSE_DEEPNESS 100
-    char          *lateClose[LATE_CLOSE_DEEPNESS];
-    int            topLateClose = 0;
-**/
     Tcl_HashEntry *h;
     domProcessingInstructionNode *pinode;
 
@@ -757,12 +733,7 @@ HTML_SimpleParse (
             if (c=='<') x++;
             while ( (c=*x)!=0 && c!='<' ) {
                 if (c=='&') ampersandSeen = 1;
-                if ( (c != ' ')  &&
-                     (c != '\t') &&
-                     (c != '\n') &&
-                     (c != '\r') ) {
-                    only_whites = 0;
-                }
+                if ( !SPACE(c) ) only_whites = 0;
                 x++;
             }
             if (!(only_whites && ignoreWhiteSpaces) && parent_node) {
@@ -773,13 +744,13 @@ HTML_SimpleParse (
                 memset(tnode, 0, sizeof(domTextNode));
                 tnode->nodeType    = TEXT_NODE;
                 tnode->nodeFlags   = 0;
-                tnode->namespace   = 0;
                 tnode->ownerDocument = doc;
                 tnode->nodeNumber  = NODE_NO(doc);
                 tnode->valueLength = (x - start);
                 tnode->nodeValue   = (char*)MALLOC((x - start)+1);
                 memmove(tnode->nodeValue, start, (x - start));
                 *(tnode->nodeValue + (x - start)) = 0;
+                DBG(fprintf(stderr, "New text node: '%s'\n", tnode->nodeValue);)
                 if (ampersandSeen) {
                     TranslateEntityRefs(tnode->nodeValue, &(tnode->valueLength) );
                 }
@@ -818,13 +789,17 @@ HTML_SimpleParse (
             \---------------------------------------------------------------------*/
             ignore = 0;
             pnode = NULL;
-            if (parent_node->lastChild && parent_node->lastChild->nodeType == ELEMENT_NODE) {
-                pnode = parent_node->lastChild;
-            } else 
-            if (parent_node->lastChild && parent_node->lastChild->previousSibling &&
-                parent_node->lastChild->previousSibling->nodeType == ELEMENT_NODE)
-            {
-                pnode = parent_node->lastChild->previousSibling;
+            if (parent_node) {
+                if (parent_node->lastChild 
+                    && (parent_node->lastChild->nodeType == ELEMENT_NODE)) {
+                    pnode = parent_node->lastChild;
+                } else 
+                    if (parent_node->lastChild 
+                        && parent_node->lastChild->previousSibling 
+                        && (parent_node->lastChild->previousSibling->nodeType 
+                            == ELEMENT_NODE)) {
+                        pnode = parent_node->lastChild->previousSibling;
+                    }
             }
             if (pnode) {
                 DBG(fprintf(stderr, "'%s' closing with last empty tag '%s' ?\n", start+2, pnode->nodeName);)
@@ -976,7 +951,6 @@ HTML_SimpleParse (
                         memset(tnode, 0, sizeof(domTextNode));
                         tnode->nodeType      = COMMENT_NODE;
                         tnode->nodeFlags     = 0;
-                        tnode->namespace     = 0;
                         tnode->ownerDocument = doc;
                         tnode->nodeNumber    = NODE_NO(doc);
                         tnode->parentNode    = parent_node;
@@ -985,16 +959,15 @@ HTML_SimpleParse (
                         memmove(tnode->nodeValue, start+4, tnode->valueLength);
                         *(tnode->nodeValue + tnode->valueLength) = 0;
                         if (parent_node == NULL) {
-                            if (doc->documentElement) {
-                                toplevel = doc->documentElement;
-                                while (toplevel->nextSibling) {
-                                    toplevel = toplevel->nextSibling;
-                                }
-                                toplevel->nextSibling   = (domNode*)tnode;
-                                tnode->previousSibling = (domNode*)toplevel;
+                            if (doc->rootNode->lastChild) {
+                                tnode->previousSibling = 
+                                    doc->rootNode->lastChild;
+                                doc->rootNode->lastChild->nextSibling 
+                                    = (domNode *) tnode;
                             } else {
-                                doc->documentElement = (domNode*)tnode;
+                                doc->rootNode->firstChild = (domNode*) tnode;
                             }
+                            doc->rootNode->lastChild = (domNode*) tnode;
                         } else {
                             if (parent_node->firstChild)  {
                                 parent_node->lastChild->nextSibling = (domNode*)tnode;
@@ -1057,7 +1030,6 @@ HTML_SimpleParse (
                             memset(tnode, 0, sizeof(domTextNode));
                             tnode->nodeType      = TEXT_NODE;
                             tnode->nodeFlags     = 0;
-                            tnode->namespace     = 0;
                             tnode->ownerDocument = doc;
                             tnode->nodeNumber    = NODE_NO(doc);
                             tnode->parentNode    = parent_node;
@@ -1133,16 +1105,14 @@ HTML_SimpleParse (
                     memmove(pinode->dataValue, piSep, pinode->dataLength);
 
                     if (parent_node == NULL) {
-                        if (doc->documentElement) {
-                            toplevel = doc->documentElement;
-                            while (toplevel->nextSibling) {
-                                toplevel = toplevel->nextSibling;
-                            }
-                            toplevel->nextSibling   = (domNode*)pinode;
-                            pinode->previousSibling = (domNode*)toplevel;
+                        if (doc->rootNode->lastChild) {
+                            pinode->previousSibling = doc->rootNode->lastChild;
+                            doc->rootNode->lastChild->nextSibling 
+                                = (domNode*) pinode;
                         } else {
-                            doc->documentElement = (domNode*)pinode;
+                            doc->rootNode->firstChild = (domNode*) pinode;
                         }
+                        doc->rootNode->lastChild = (domNode*) pinode;
                     } else {
                         if (parent_node->firstChild)  {
                             parent_node->lastChild->nextSibling = (domNode*)pinode;
@@ -1165,7 +1135,7 @@ HTML_SimpleParse (
             |
             \---------------------------------------------------------------*/
             while ((c=*x)!=0 && c!='/' && c!='>' && c!='<' && !SPACE(c) ) {
-                if (!isalpha(c)) goto readText;
+                if (!isalnum(c)) goto readText;
                 *x = tolower(c);
                 x++;
             }
@@ -1192,7 +1162,9 @@ HTML_SimpleParse (
                           break;
                 case 'b': if(!strcmp(e,"b")&&!strcmp(pn,"b")) autoclose=1;
                           break;
-                case 'p': if(!strcmp(e,"pre")&&!strcmp(pn,"pre")) autoclose=1;
+                case 'o': if (!strcmp(e,"option")&&!strcmp(pn,"option")) autoclose = 1; break;
+                case 'p': if((!strcmp(e,"pre")&&!strcmp(pn,"pre")) 
+                             ||(!strcmp(e,"p")&&!strcmp(pn,"p"))) autoclose=1;
                           break;
             }
             if (autoclose) {
@@ -1204,7 +1176,7 @@ HTML_SimpleParse (
             /*-----------------------------------------------------------
             |   create new DOM element node
             \----------------------------------------------------------*/
-            h = Tcl_CreateHashEntry(&HASHTAB(doc,tagNames), e, &hnew);
+            h = Tcl_CreateHashEntry(&HASHTAB(doc,tdom_tagNames), e, &hnew);
 
             node = (domNode*) domAlloc(sizeof(domNode));
             memset(node, 0, sizeof(domNode));
@@ -1216,15 +1188,13 @@ HTML_SimpleParse (
             node->nodeNumber    = NODE_NO(doc);
 
             if (parent_node == NULL) {
-                if (doc->documentElement) {
-                    toplevel = doc->documentElement;
-                    while (toplevel->nextSibling) {
-                        toplevel = toplevel->nextSibling;
-                    }
-                    toplevel->nextSibling = node;
-                    node->previousSibling = toplevel;
+                if (doc->rootNode->lastChild) {
+                    node->previousSibling = doc->rootNode->lastChild;
+                    doc->rootNode->lastChild->nextSibling = node;
+                } else {
+                    doc->rootNode->firstChild = node;
                 }
-                doc->documentElement = node;
+                doc->rootNode->lastChild = node;
             } else {
                 node->parentNode = parent_node;
                 if (parent_node->firstChild)  {
@@ -1252,15 +1222,21 @@ HTML_SimpleParse (
                 int nArgVal = 0;
 
                 while ((c=*x)!=0 && c!='=' && c!='>' && !SPACE(c) ) {
+                    *x = tolower(c);
                     x++;
                 }
                 nArgName = x - ArgName;
-                
+                while (SPACE(*x)) {
+                    x++;
+                }
                 if (*x == '=') {
                 
                     /* attribute with value, like width="1234" */
                     
                     x++;
+                    while (SPACE(*x)) {
+                        x++;
+                    }
                     saved = *(ArgName + nArgName);
                     *(ArgName + nArgName) = '\0'; /* terminate arg name */
 
@@ -1295,17 +1271,17 @@ HTML_SimpleParse (
                     }
                 } else {
                     /* attribute without value, like 'nowrap' */
-                    x++;
                     saved = *(ArgName + nArgName);
                     *(ArgName + nArgName) = '\0'; /* terminate arg name */
-                    ArgVal = "1"; /* current hack */
-                    nArgVal = 1;                
+                    ArgVal = ArgName;
+                    nArgVal = nArgName;
                 }
 
                 /*--------------------------------------------------
                 |   allocate new attribute node
                 \--------------------------------------------------*/
-                h = Tcl_CreateHashEntry(&HASHTAB(doc,attrNames), ArgName, &hnew);
+                h = Tcl_CreateHashEntry(&HASHTAB(doc,tdom_attrNames),
+                                        ArgName, &hnew);
                 attrnode = (domAttrNode*) domAlloc(sizeof(domAttrNode));
                 memset(attrnode, 0, sizeof(domAttrNode));
                 attrnode->parentNode  = node;
@@ -1317,7 +1293,24 @@ HTML_SimpleParse (
                 memmove(attrnode->nodeValue, ArgVal, nArgVal);
                 *(attrnode->nodeValue + nArgVal) = 0;
                 if (ampersandSeen) {
-                    TranslateEntityRefs(attrnode->nodeValue, &(attrnode->valueLength) );
+                    TranslateEntityRefs(attrnode->nodeValue, 
+                                        &(attrnode->valueLength) );
+                }
+                if (!strcmp(ArgName, "id")) {
+                    if (!doc->ids) {
+                        doc->ids = 
+                            (Tcl_HashTable *) MALLOC (sizeof (Tcl_HashTable));
+                        Tcl_InitHashTable (doc->ids, TCL_STRING_KEYS);
+                    }
+                    h = Tcl_CreateHashEntry (doc->ids, attrnode->nodeValue,
+                                             &hnew);
+                    /* How to resolve in case of dublicates?  We
+                       follow, what the core dom building code does:
+                       the first value in document order wins. */
+                    if (hnew) {
+                        Tcl_SetHashValue (h, node);
+                        attrnode->nodeFlags |= IS_ID_ATTRIBUTE;
+                    }
                 }
                 if (node->firstAttr) {
                     lastAttr->nextSibling = attrnode;
@@ -1349,10 +1342,10 @@ HTML_SimpleParse (
                                !strcmp(node->nodeName,"input") ||
                                !strcmp(node->nodeName,"isindex"))  hasContent = 0; break;
                 case 'l':  if (!strcmp(node->nodeName,"link"))     hasContent = 0; break;
-                case 'o':  if (!strcmp(node->nodeName,"option"))   hasContent = 0; break;
                 case 'm':  if (!strcmp(node->nodeName,"meta"))     hasContent = 0; break;
                 case 'p':  if (!strcmp(node->nodeName,"param"))    hasContent = 0; break;
-                case 's':  if (!strcmp(node->nodeName,"spacer"))   hasContent = 0; break; /*ext*/
+                case 's':  if (!strcmp(node->nodeName,"spacer") ||                        /*ext*/
+                               !strcmp(node->nodeName,"source"))   hasContent = 0; break; /*html5*/
             }
 
             if (*x=='/') {
@@ -1404,7 +1397,6 @@ HTML_SimpleParse (
                     memset(tnode, 0, sizeof(domTextNode));
                     tnode->nodeType      = TEXT_NODE;
                     tnode->nodeFlags     = 0;
-                    tnode->namespace     = 0;
                     tnode->ownerDocument = doc;
                     tnode->nodeNumber    = NODE_NO(doc);
                     tnode->parentNode    = node;
@@ -1431,14 +1423,13 @@ HTML_SimpleParse (
         }
     }
 
-    while (parent_node != NULL) {
-
+    while (parent_node != NULL && node->parentNode) {
         pn = (char*)node->parentNode->nodeName;
         DBG(fprintf(stderr, "final autoclose '%s'? \n", pn);)
         /*---------------------------------------------------------------
         |   check for tags for which end tag can be omitted
         \--------------------------------------------------------------*/
-        autoclose = 1;
+        autoclose = 0;
         switch (pn[0]) {
             case 'b': if (!strcmp(pn,"body"))     autoclose = 1; break;
             case 'c': if (!strcmp(pn,"colgroup")) autoclose = 1; break;
@@ -1485,27 +1476,12 @@ HTML_SimpleParseDocument (
     int    *pos,
     char  **errStr
 ) {
-    Tcl_HashEntry *h;
-    domNode       *rootNode;
-    int            hnew;
-    domDocument   *doc = domCreateEmptyDoc();
+    domDocument   *doc = domCreateDoc(NULL, 0);
 
     *pos = 0;
     HTML_SimpleParse (html, pos, doc, NULL, ignoreWhiteSpaces, errStr);
 
-    h = Tcl_CreateHashEntry(&HASHTAB(doc,tagNames), "(rootNode)", &hnew);
-    rootNode = (domNode*) domAlloc(sizeof(domNode));
-
-    memset(rootNode, 0, sizeof(domNode));
-    rootNode->nodeType      = ELEMENT_NODE;
-    rootNode->nodeFlags     = 0;
-    rootNode->namespace     = 0;
-    rootNode->nodeName      = (char *)&(h->key);
-    rootNode->ownerDocument = doc;
-    rootNode->nodeNumber    = NODE_NO(doc);
-    rootNode->parentNode    = NULL;
-    rootNode->firstChild = rootNode->lastChild = doc->documentElement;
-    doc->rootNode = rootNode;
+    domSetDocumentElement (doc);
     return doc;
 
 } /* HTML_SimpleParseDocument */
