@@ -33,28 +33,9 @@
 |
 |   ------------------------------------------------------------------------
 |
-|   A parser for XML.
-|
-|   Copyright (C) 1998 D. Richard Hipp
-|
-|   This library is free software; you can redistribute it and/or
-|   modify it under the terms of the GNU Library General Public
-|   License as published by the Free Software Foundation; either
-|   version 2 of the License, or (at your option) any later version.
-|
-|   This library is distributed in the hope that it will be useful,
-|   but WITHOUT ANY WARRANTY; without even the implied warranty of
-|   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-|   Library General Public License for more details.
-|
-|   You should have received a copy of the GNU Library General Public
-|   License along with this library; if not, write to the
-|   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-|   Boston, MA  02111-1307, USA.
-|
-|   Author contact information:
-|     drh@acm.org
-|     http://www.hwaci.com/drh/
+|   Partly based on a parser for XML (for TMML by R.Hipp 1998). 
+|   This source code is released into the public domain by the author.
+|   on 2002, December 17.
 |
 \---------------------------------------------------------------------------*/
 
@@ -208,7 +189,7 @@ static Er er_sequences[] = {
 \---------------------------------------------------------------------------*/
 static void ErInit (void)
 {
-    int i;  /* For looping thru the list of entity references */
+    size_t i;  /* For looping thru the list of entity references */
     int h;  /* The hash on a entity */
 
     for(i=0; i<sizeof(er_sequences)/sizeof(er_sequences[0]); i++){
@@ -234,7 +215,7 @@ static void ErInit (void)
 |          output =   "AT&T > MCI"
 |
 \---------------------------------------------------------------------------*/
-static void TranslateEntityRefs (
+static int TranslateEntityRefs (
     char *z,
     int  *newLen
 )
@@ -273,7 +254,7 @@ static void TranslateEntityRefs (
                 value = 0;
                 if (z[++i] == 'x') {
                     i++;
-                    while ((c=z[i]) && (c!=';')) {
+                    while (z[i] && (c=z[i]) && (c!=';')) {
                         value = value * 16;
                         if ((c>='0') && (c<='9')) {
                             value += c-'0';
@@ -285,21 +266,24 @@ static void TranslateEntityRefs (
                             value += c-'a' + 10;
                         } else {
                             /* error */
+                            return 0;
                         }
                         i++;
                     }
                 } else {
-                    while ((c=z[i]) && (c!=';')) {
+                    while (z[i] && (c=z[i]) && (c!=';')) {
                         value = value * 10;
                         if ((c>='0') && (c<='9')) {
                             value += c-'0';
                         } else {
                             /* error */
+                            return 0;
                         }
                         i++;
                     }
                 }
-                if (z[i]!=';') {
+                if (!z[i] || (z[i]!=';')) {
+                    return 0;
                     /* error */
                 }
                 from = i+1;
@@ -317,12 +301,16 @@ static void TranslateEntityRefs (
                     z[to++] = (char) ((value | 0x80) & 0xBF);
                 } else {
                     /* error */
+                    return 0;
                 }
 #endif
             } else {
-                while (z[i] && isalpha(z[i])) {
+                while (z[i] && isalpha((unsigned char)z[i])) {
                    i++;
                 }
+                if (!z[i] || (z[i]!=';')) {
+                    return 0;
+                }
                 c = z[i];
                 z[i] = 0;
                 h = ErHash(&z[from+1]);
@@ -348,6 +336,7 @@ static void TranslateEntityRefs (
     }
     z[to] = 0;
     *newLen = to;
+    return 1;
 }
 /*----------------------------------------------------------------------------
 |   End Of Character Entity Translator
@@ -362,8 +351,8 @@ static int
 domIsNamespaceInScope (
     domActiveNS *NSstack,
     int          NSstackPos,
-    char        *prefix,
-    char        *namespaceURI
+    const char  *prefix,
+    const char  *namespaceURI
 )
 {
     int    i;
@@ -398,7 +387,6 @@ XML_SimpleParse (
     domDocument *doc,
     domNode     *parent_nodeOld,
     int          ignoreWhiteSpaces,
-    char        *baseURI,
     char       **errStr
 ) {
     register int   c;          /* Next character of the input file */
@@ -406,7 +394,7 @@ XML_SimpleParse (
     register char *x, *start, *piSep;
     int            saved;
     int            hasContent;
-    domNode       *node, *toplevel, *rootNode;
+    domNode       *node;
     domNode       *parent_node = NULL;
     domTextNode   *tnode;
     domAttrNode   *attrnode, *lastAttr, *attrList;
@@ -422,7 +410,7 @@ XML_SimpleParse (
     int            activeNSpos  = -1;
     int            activeNSsize = 8;
     domActiveNS   *activeNS     = (domActiveNS*) MALLOC (sizeof(domActiveNS) * activeNSsize);
-    char          *xmlns, *localname;
+    const char    *xmlns, *localname;
     domNS         *ns;
     char           tagPrefix[MAX_PREFIX_LEN];
     char           prefix[MAX_PREFIX_LEN];
@@ -460,7 +448,6 @@ XML_SimpleParse (
                 memset(tnode, 0, sizeof(domTextNode));
                 tnode->nodeType    = TEXT_NODE;
                 tnode->nodeFlags   = 0;
-                tnode->namespace   = 0;
                 tnode->ownerDocument = doc;
                 tnode->nodeNumber  = NODE_NO(doc);
                 tnode->valueLength = (x - start);
@@ -468,7 +455,10 @@ XML_SimpleParse (
                 memmove(tnode->nodeValue, start, (x - start));
                 *(tnode->nodeValue + (x - start)) = 0;
                 if (ampersandSeen) {
-                    TranslateEntityRefs(tnode->nodeValue, &(tnode->valueLength) );
+                    if (!TranslateEntityRefs(tnode->nodeValue, 
+                                             &(tnode->valueLength) )) {
+                        RetError("Entity parsing error", (x - xml));
+                    }
                 }
                 tnode->parentNode = parent_node;
                 if (parent_node->firstChild)  {
@@ -476,7 +466,8 @@ XML_SimpleParse (
                     tnode->previousSibling = parent_node->lastChild;
                     parent_node->lastChild = (domNode*)tnode;
                 } else {
-                    parent_node->firstChild = parent_node->lastChild = (domNode*)tnode;
+                    parent_node->firstChild = parent_node->lastChild = 
+                        (domNode*)tnode;
                 }
             }
 
@@ -485,6 +476,9 @@ XML_SimpleParse (
             |   read and check closing tag
             \-----------------------------------------------------------*/
             node = parent_node;
+            if (!parent_node) {
+                RetError("Syntax error",(x - xml));
+            }
             parent_node = node->parentNode;
             pn = (char*)node->nodeName;
 
@@ -511,20 +505,7 @@ XML_SimpleParse (
             } 
 #endif                                   
             if (parent_node == NULL) {
-                /* we return to main node and so finished parsing
-                   create the root node now
-                 */
-                rootNode = doc->rootNode;
-                rootNode->firstChild = doc->documentElement;
-                while (rootNode->firstChild->previousSibling) {
-                    rootNode->firstChild = rootNode->firstChild->previousSibling;
-                }
-                rootNode->lastChild = doc->documentElement;
-                while (rootNode->lastChild->nextSibling) {
-                    rootNode->lastChild = rootNode->lastChild->nextSibling;
-                }
-                doc->rootNode = rootNode;
-
+                /* we return to main node and so finished parsing */
 #ifdef TDOM_NS
                 FREE ((char *) activeNS);
 #endif                
@@ -553,7 +534,6 @@ XML_SimpleParse (
                         memset(tnode, 0, sizeof(domTextNode));
                         tnode->nodeType      = COMMENT_NODE;
                         tnode->nodeFlags     = 0;
-                        tnode->namespace     = 0;
                         tnode->ownerDocument = doc;
                         tnode->nodeNumber    = NODE_NO(doc);
                         tnode->parentNode    = parent_node;
@@ -562,16 +542,15 @@ XML_SimpleParse (
                         memmove(tnode->nodeValue, start+4, tnode->valueLength);
                         *(tnode->nodeValue + tnode->valueLength) = 0;
                         if (parent_node == NULL) {
-                            if (doc->documentElement) {
-                                toplevel = doc->documentElement;
-                                while (toplevel->nextSibling) {
-                                    toplevel = toplevel->nextSibling;
-                                }
-                                toplevel->nextSibling   = (domNode*)tnode;
-                                tnode->previousSibling = (domNode*)toplevel;
+                            if (doc->rootNode->lastChild) {
+                                tnode->previousSibling = 
+                                    doc->rootNode->lastChild;
+                                doc->rootNode->lastChild->nextSibling 
+                                    = (domNode*)tnode;
                             } else {
-                                doc->documentElement = (domNode*)tnode;
+                                doc->rootNode->firstChild = (domNode*) tnode;
                             }
+                            doc->rootNode->lastChild = (domNode*) tnode;
                         } else {
                             if (parent_node->firstChild)  {
                                 parent_node->lastChild->nextSibling = (domNode*)tnode;
@@ -626,7 +605,7 @@ XML_SimpleParse (
                         x++;
                     }
                     if (*x) {
-                        if (parent_node) {
+                        if (parent_node && (x - start)) {
                             /*----------------------------------------------------
                             |   allocate new TEXT node for CDATA section data
                             \---------------------------------------------------*/
@@ -634,7 +613,6 @@ XML_SimpleParse (
                             memset(tnode, 0, sizeof(domTextNode));
                             tnode->nodeType      = TEXT_NODE;
                             tnode->nodeFlags     = 0;
-                            tnode->namespace     = 0;
                             tnode->ownerDocument = doc;
                             tnode->nodeNumber    = NODE_NO(doc);
                             tnode->parentNode    = parent_node;
@@ -710,16 +688,14 @@ XML_SimpleParse (
                     memmove(pinode->dataValue, piSep, pinode->dataLength);
 
                     if (parent_node == NULL) {
-                        if (doc->documentElement) {
-                            toplevel = doc->documentElement;
-                            while (toplevel->nextSibling) {
-                                toplevel = toplevel->nextSibling;
-                            }
-                            toplevel->nextSibling   = (domNode*)pinode;
-                            pinode->previousSibling = (domNode*)toplevel;
+                        if (doc->rootNode->lastChild) {
+                            pinode->previousSibling = doc->rootNode->lastChild;
+                            doc->rootNode->lastChild->nextSibling 
+                                = (domNode*) pinode;
                         } else {
-                            doc->documentElement = (domNode*)pinode;
+                            doc->rootNode->firstChild = (domNode*) pinode;
                         }
+                        doc->rootNode->lastChild = (domNode*) pinode;
                     } else {
                         if (parent_node->firstChild)  {
                             parent_node->lastChild->nextSibling = (domNode*)pinode;
@@ -755,7 +731,8 @@ XML_SimpleParse (
             /*------------------------------------------------------
             |   create new DOM element node
             \-----------------------------------------------------*/
-            h = Tcl_CreateHashEntry(&HASHTAB(doc,tagNames), start+1, &hnew);
+            h = Tcl_CreateHashEntry(&HASHTAB(doc,tdom_tagNames), start+1,
+                                    &hnew);
             node = (domNode*) domAlloc(sizeof(domNode));
             memset(node, 0, sizeof(domNode));
             node->nodeType      = ELEMENT_NODE;
@@ -767,15 +744,13 @@ XML_SimpleParse (
             node->ownerDocument = doc;
 
             if (parent_node == NULL) {
-                if (doc->documentElement) {
-                    toplevel = doc->documentElement;
-                    while (toplevel->nextSibling) {
-                        toplevel = toplevel->nextSibling;
-                    }
-                    toplevel->nextSibling = node;
-                    node->previousSibling = toplevel;
+                if (doc->rootNode->lastChild) {
+                    node->previousSibling = doc->rootNode->lastChild;
+                    doc->rootNode->lastChild->nextSibling = node;
+                } else {
+                    doc->rootNode->firstChild = node;
                 }
-                doc->documentElement = node;
+                doc->rootNode->lastChild = node;
             } else {
                 node->parentNode = parent_node;
                 if (parent_node->firstChild)  {
@@ -866,7 +841,8 @@ XML_SimpleParse (
                     xmlns = ArgName;
                     newNS = 1;
                     
-                    h = Tcl_CreateHashEntry(&HASHTAB(doc, attrNames),ArgName, &hnew); 
+                    h = Tcl_CreateHashEntry(&HASHTAB(doc, tdom_attrNames),
+                                            ArgName, &hnew); 
                     attrnode = (domAttrNode*) domAlloc(sizeof(domAttrNode));
                     memset(attrnode, 0, sizeof(domAttrNode));
                     attrnode->parentNode  = node;
@@ -878,8 +854,10 @@ XML_SimpleParse (
                     memmove(attrnode->nodeValue, ArgVal, nArgVal);
                     *(attrnode->nodeValue + nArgVal) = 0;
                     if (ampersandSeen) {
-                        TranslateEntityRefs(attrnode->nodeValue,
-                                            &(attrnode->valueLength) );
+                        if (!TranslateEntityRefs(attrnode->nodeValue,
+                                                 &(attrnode->valueLength) )) {
+                            RetError("Entity parsing error",(start-xml));
+                        }
                     }
                     
                     if (xmlns[5] == ':') {
@@ -925,7 +903,8 @@ XML_SimpleParse (
                     /*------------------------------------------------------------
                     |   allocate new attribute node
                     \------------------------------------------------------------*/
-                    h = Tcl_CreateHashEntry(&HASHTAB(doc,attrNames), ArgName, &hnew);
+                    h = Tcl_CreateHashEntry(&HASHTAB(doc,tdom_attrNames),
+                                            ArgName, &hnew);
                     attrnode = (domAttrNode*) domAlloc(sizeof(domAttrNode));
                     memset(attrnode, 0, sizeof(domAttrNode));
                     attrnode->parentNode  = node;
@@ -937,7 +916,10 @@ XML_SimpleParse (
                     memmove(attrnode->nodeValue, ArgVal, nArgVal);
                     *(attrnode->nodeValue + nArgVal) = 0;
                     if (ampersandSeen) {
-                        TranslateEntityRefs(attrnode->nodeValue, &(attrnode->valueLength) );
+                        if (!TranslateEntityRefs(attrnode->nodeValue,
+                                                 &(attrnode->valueLength) )) {
+                            RetError("Entity parsing error", (start - xml));
+                        }
                     }
                     if (attrList) {
                         lastAttr->nextSibling = attrnode;
@@ -958,7 +940,7 @@ XML_SimpleParse (
             /*----------------------------------------------------------
             |   look for namespace of element
             \---------------------------------------------------------*/
-            domSplitQName ((char*)node->nodeName, tagPrefix,
+            domSplitQName (node->nodeName, tagPrefix,
                            &localname);
             for (nspos = activeNSpos; nspos >= 0; nspos--) {
                 if (  ((tagPrefix[0] == '\0') && (activeNS[nspos].namespace->prefix[0] == '\0'))
@@ -1024,18 +1006,6 @@ XML_SimpleParse (
                 }
             }
             if (x[1] == 0) {
-
-                rootNode = doc->rootNode;
-                rootNode->firstChild = doc->documentElement;
-                while (rootNode->firstChild->previousSibling) {
-                    rootNode->firstChild = rootNode->firstChild->previousSibling;
-                }
-                rootNode->lastChild = doc->documentElement;
-                while (rootNode->lastChild->nextSibling) {
-                    rootNode->lastChild = rootNode->lastChild->nextSibling;
-                }
-                doc->rootNode = rootNode;
-
 #ifdef TDOM_NS
                 FREE ((char *) activeNS);
 #endif                
@@ -1077,40 +1047,15 @@ XML_SimpleParseDocument (
     int     *pos,
     char   **errStr
 ) {
-    int            hnew;
-    Tcl_HashEntry *h;
-    domNode       *rootNode;
-    domDocument   *doc = domCreateEmptyDoc();
+    domDocument   *doc = domCreateDoc(baseURI, 0);
 
     if (extResolver) {
-        doc->extResolver = extResolver;
-        Tcl_IncrRefCount(extResolver);
+        doc->extResolver = tdomstrdup (Tcl_GetString (extResolver));
     }
     
-    h = Tcl_CreateHashEntry(&HASHTAB(doc,tagNames), "(rootNode)", &hnew);
-    rootNode = (domNode*) domAlloc(sizeof(domNode));
-    memset(rootNode, 0, sizeof(domNode));
-    rootNode->nodeType      = ELEMENT_NODE;
-    rootNode->nodeFlags     = 0;
-    rootNode->namespace     = 0;
-    rootNode->nodeName      = (char *)&(h->key);
-    rootNode->ownerDocument = doc;
-    rootNode->nodeNumber    = NODE_NO(doc);
-    rootNode->parentNode    = NULL;
-#ifdef TDOM_NS
-    rootNode->firstAttr     = domCreateXMLNamespaceNode (rootNode);
-#endif    
-    if (baseURI) {
-        h = Tcl_CreateHashEntry (&doc->baseURIs, 
-                                 (char*)rootNode->nodeNumber,
-                                 &hnew);
-        Tcl_SetHashValue (h, tdomstrdup (baseURI));
-                    rootNode->nodeFlags |= HAS_BASEURI;
-    }
-    doc->rootNode = rootNode;
-    
     *pos = 0;
-    XML_SimpleParse (xml, pos, doc, NULL, ignoreWhiteSpaces, baseURI, errStr);
+    XML_SimpleParse (xml, pos, doc, NULL, ignoreWhiteSpaces, errStr);
+    domSetDocumentElement (doc);
 
     return doc;
 
