@@ -2,7 +2,7 @@
 #   Copyright (c) 1999 Jochen Loewer (loewerj@hotmail.com)
 #----------------------------------------------------------------------------
 #
-#   $Header: /usr/local/pubcvs/tdom/lib/tdom.tcl,v 1.14.2.1 2003/07/06 22:43:54 rolf Exp $
+#   $Id$
 #
 #
 #   The higher level functions of tDOM written in plain Tcl.
@@ -54,6 +54,8 @@ namespace eval ::dom {
 namespace eval ::tDOM { 
     variable extRefHandlerDebug 0
     variable useForeignDTD ""
+
+    namespace export xmlOpenFile xmlReadFile extRefHandler baseURL
 }
 
 #----------------------------------------------------------------------------
@@ -154,7 +156,7 @@ proc ::dom::domDoc::info { doc subcommand args } {
 #----------------------------------------------------------------------------
 proc ::dom::domDoc::importNode { doc importedNode deep } {
 
-    if {$deep || ($deep == "-deep"} {
+    if {$deep || ($deep == "-deep")} {
         set node [$importedNode cloneNode -deep]
     } else {
         set node [$importedNode cloneNode]
@@ -334,7 +336,7 @@ proc ::dom::domNode::item { nodeListNode index } {
 #----------------------------------------------------------------------------
 proc ::dom::domNode::length { nodeListNode } {
 
-    return [llength [$nodeListNode childNodes] $childNodes]
+    return [llength [$nodeListNode childNodes]]
 }
 
 #----------------------------------------------------------------------------
@@ -537,20 +539,33 @@ proc ::dom::xpathFunc::element-available { ctxNode pos
     }
     foreach { arg1Typ arg1Value } $args break
     set str [::dom::xpathFuncHelper::coerce2string $arg1Typ $arg1Value ]
+    # The XSLT recommendation says: "The element-available
+    # function returns true if and only if the expanded-name
+    # is the name of an instruction." The following xsl
+    # elements are not in the category instruction.
+    # xsl:attribute-set 
+    # xsl:decimal-format 
+    # xsl:include
+    # xsl:key 
+    # xsl:namespace-alias
+    # xsl:output
+    # xsl:param
+    # xsl:strip-space
+    # xsl:preserve-space
+    # xsl:template
+    # xsl:import
+    # xsl:otherwise
+    # xsl:sort
+    # xsl:stylesheet
+    # xsl:transform
+    # xsl:with-param
+    # xsl:when
     switch $str {
-        xsl:stylesheet -
-        xsl:transform -
-        xsl:include -
-        xsl:import -
-        xsl:strip-space -
-        xsl:preserve-space -
-        xsl:template -
         xsl:apply-templates -
         xsl:apply-imports -
         xsl:call-template -
         xsl:element -
         xsl:attribute -
-        xsl:attribute-set -
         xsl:text -
         xsl:processing-instruction -
         xsl:comment -
@@ -560,21 +575,12 @@ proc ::dom::xpathFunc::element-available { ctxNode pos
         xsl:for-each -
         xsl:if -
         xsl:choose -
-        xsl:when -
-        xsl:otherwise -
-        xsl:sort -
         xsl:variable -
-        xsl:param -
         xsl:copy-of -
-        xsl:with-param -
-        xsl:key -
         xsl:message -
-        xsl:decimal-format -
-        xsl:namespace-alias -
         xsl:fallback {
             return [list bool true]
         }
-        xsl:output -
         default {
             return [list bool false]
         }
@@ -675,10 +681,19 @@ proc tDOM::IANAEncoding2TclEncoding {IANAName} {
         "gb2312"      {return gb2312}
         "big5"        {return big5}
         "cp866"       {return cp866}
+        "cp1250"      {return cp1250}
+        "cp1253"      {return cp1253}
+        "cp1254"      {return cp1254}
+        "cp1255"      {return cp1255}
+        "cp1256"      {return cp1256}
+        "cp1257"      {return cp1257}
 
         "windows-1251" -
         "cp1251"      {return cp1251}
-        
+
+        "windows-1252" -
+        "cp1252"      {return cp1252}    
+
         "iso_8859-1:1987" -
         "iso-ir-100" -
         "iso_8859-1" -
@@ -774,13 +789,13 @@ proc tDOM::xmlOpenFile {filename {encodingString {}}} {
             # Try to find the end of the XML Declaration
             set closeIndex [string first ">" $head]
             if {$closeIndex == -1} {
-                error "Wired XML data or not XML data at all"
+                error "Weird XML data or not XML data at all"
             }
 
             seek $fd 0 start
             set xmlDeclaration [read $fd [expr {$closeIndex + 5}]]
             # extract the encoding information
-            set pattern {encoding=[\x20\x9\xd\xa]*["']([^ "']+)['"]}
+            set pattern {^[^>]+encoding=[\x20\x9\xd\xa]*["']([^ "']+)['"]}
             # emacs: "
             if {![regexp $pattern $head - encStr]} {
                 # Probably something like <?xml version="1.0"?>. 
